#include<avr/io.h>

#define uint8 unsigned char
#define uint16 unsigned int
#define uint32 unsigned long

#define DLY	\
asm ("NOP"); \
asm ("NOP");

// control line pins
#define	IM0		1
#define	IM1		2
#define	IM2		4
#define	RES		8
#define	CS		0x10
#define	CD		0x20
#define	WR		0x40
#define	RD		0x80

#define OUT_LO	PORTE
#define IN_LO	PINE
#define DDR_LO	DDRE

#define OUT_HI	PORTA
#define IN_HI	PINA
#define DDR_HI	DDRA

#define CTRL	PORTC
#define DDR_CT	DDRC

#define MODE16	1 //IM1
#define MODE8	3 //(IM1 + IM0)

#define OUTS	0xFF
#define INS		0

// macros for control lines
#define CLR_CD 		CTRL &= ~CD;
#define SET_CD 		CTRL |= CD;

#define CLR_CS 		CTRL &= ~CS;
#define SET_CS 		CTRL |= CS;

#define CLR_RESET 	CTRL &= ~RES;
#define SET_RESET 	CTRL |= RES;

#define CLR_WR 		CTRL &= ~WR;
#define SET_WR 		CTRL |= WR;

#define CLR_RD 		CTRL &= ~RD;
#define SET_RD 		CTRL |= RD;

// color definitions
// maximum values for R, G, B are 31, 63, 31 (5b, 6b, 5b)
// 16 bit color RRRR RGGG GGGB BBBB
#define RGB(r,g,b)	(uint16)(((b & 0x1F) * 2048) | ((g & 0x3F) * 32) | (r & 0x1F))

#define	BLACK	0x0000
#define	BLUE	0xF800 //RGB(0,0,31) //0x001F
#define	RED		0x001F //RGB(31,0,0) //0xF800
#define	GREEN	0x07E0 //RGB(0,63,0) //0x07E0
#define CYAN	0xFFE0 //RGB(0,63,31) //0x07FF
#define MAGENTA	0xF81F //RGB(31,0,31) //0xF81F
#define YELLOW	0x07FF //RGB(31,63,0) //0xFFE0
#define WHITE	0xFFFF //RGB(31,63,31) //0xFFFF

uint8	mode;

/*************************************************/
void delay(uint16 t)
{
	uint8 t1;
	while(t--)
	{
		for(t1=11; t1>0; t1--)
		{
			asm("nop");
		}
	}
}

/*************************************************/
void write_command(uint16 arg)
{
	CLR_CS;
	CLR_CD;

	if (mode == 16)
	{
		OUT_HI = arg >> 8;
		OUT_LO = arg & 0xFF;
		DLY
		CLR_WR;
		DLY
		SET_WR;
	}
	else
	{
		OUT_HI = arg >> 8;
		CLR_WR;
		DLY
		SET_WR;

		OUT_HI = arg & 0xFF;
		CLR_WR;
		DLY
		SET_WR;
	}
	SET_CS;
}

/*************************************************/
void write_data(uint16 arg)
{
	CLR_CS;
	SET_CD;
	
	if (mode == 16)
	{
		OUT_HI = arg >> 8;
		OUT_LO = arg & 0xFF;
		DLY
		CLR_WR;
		DLY
		SET_WR;
	}
	else
	{
		OUT_HI = arg >> 8;
		CLR_WR;
		DLY
		SET_WR;

		OUT_HI = arg & 0xFF;
		CLR_WR;
		DLY
		SET_WR;
	}
	SET_CS;
}

/*************************************************/
uint16 LCDRead(void)
{
	uint8 argh, argl;

	CLR_CS;
	SET_CD;
	DDR_HI = INS;	// set hi port to inputs
	if (mode == 16)
	{
		DDR_LO = INS;	// set lo port to inputs
		CLR_RD;			// assert read
		DLY;
		DLY;
		DLY;
		argh = IN_HI;	// get word from the bus
		argl = IN_LO;
		SET_RD;			// de-assert read
		DDR_LO = OUTS;	// set lo port to outputs
	}
	else
	{
		CLR_RD;			// assert read
		DLY;
		DLY;
		DLY;
		argh = IN_HI;	// get byte from the bus
		SET_RD;			// de-assert read
		DLY;
		DLY;
		CLR_RD;			// assert read
		DLY;
		DLY;
		DLY;
		argl = IN_HI;	// get byte from the bus
		SET_RD;			// de-assert read
	}
	DDR_HI = OUTS;	// set hi port to outputs
	SET_CS;
	return (uint16)((argh << 8) | argl);
}

/*************************************************/
void init()
{
	write_command(0x0606);	// Pin Control (R606h)
	write_data(0x0000);		// Page 41 of SPFD5420A Datasheet
	delay(50);
	write_command(0x0007);	// Display Control 1 (R007h)
	write_data(0x0001);		// Page 16 of SPFD5420A Datasheet 
	delay(50);
	write_command(0x0110);	// Power Control 6(R110h)
	write_data(0x0001);		// Page 30 of SPFD5420A Datasheet
	delay(50);
	write_command(0x0100);	// Power Control 1 (R100h)
	write_data(0x17B0);		// Page 26 of SPFD5420A Datasheet
	write_command(0x0101);	// Power Control 2 (R101h)
	write_data(0x0147);		// Page 27 of SPFD5420A Datasheet
	write_command(0x0102);	// Power Control 3 (R102h)
	write_data(0x019D);		// Page 28 of SPFD5420A Datasheet
	write_command(0x0103);	// Power Control 4 (R103h)
	write_data(0x3600);		// Page 29 of SPFD5420A Datasheet
	write_command(0x0281);	// NVM read data 2 (R281h)  
	write_data(0x0010);		// Page 34 of SPFD5420A Datasheet
	delay(50);
	write_command(0x0102);	// Power Control 3 (R102h)
	write_data(0x01BD);		// Page 28 of SPFD5420A Datasheet
	delay(50);

	//--------------- Power control 1~6 ---------------//
	write_command(0x0100);	// Power Control 1 (R100h)
	write_data(0x16B0);		// Page 26 of SPFD5420A Datasheet
	write_command(0x0101);	// Power Control 2 (R101h)
	write_data(0x0147);		// Page 27 of SPFD5420A Datasheet
	write_command(0x0102);	// Power Control 3 (R102h)
	write_data(0x01BD);		// Page 28 of SPFD5420A Datasheet
	write_command(0x0103);	// Power Control 4 (R103h)
	write_data(0x2d00);		// Page 29 of SPFD5420A Datasheet
	write_command(0x0107);	// Power Control 5 (R107h)
	write_data(0x0000);		// Page 30 of SPFD5420A Datasheet
	write_command(0x0110);	// Power Control 6(R110h)
	write_data(0x0001);		// Page 30 of SPFD5420A Datasheet
	write_command(0x0280);	// NVM read data 1 (R280h)
	write_data(0x0000);		// Page 33 of SPFD5420A Datasheet
	write_command(0x0281);	// NVM read data 2 (R281h)
	write_data(0x0006);		// Page 34 of SPFD5420A Datasheet
	write_command(0x0282);	// NVM read data 3 (R282h)
	write_data(0x0000);		// Page 34 of SPFD5420A Datasheet

	//------- Gamma 2.2 control (R300h to R30Fh) ------//
/*	write_command(0x0300);
	write_data(0x0101);
	write_command(0x0301);
	write_data(0x0b27);
	write_command(0x0302);
	write_data(0x132a);
	write_command(0x0303);
	write_data(0x2a13);
	write_command(0x0304);
	write_data(0x270b);
	write_command(0x0305);
	write_data(0x0101);
	write_command(0x0306);
	write_data(0x1205);
	write_command(0x0307);
	write_data(0x0512);
	write_command(0x0308);
	write_data(0x0005);
	write_command(0x0309);
	write_data(0x0003);
	write_command(0x030A);
	write_data(0x0f04);
	write_command(0x030B);
	write_data(0x0f00);
	write_command(0x030C);
	write_data(0x000f);
	write_command(0x030D);
	write_data(0x040f);
	write_command(0x030E);
	write_data(0x0300);
	write_command(0x030F);
	write_data(0x0500);*/

	write_command(0x0400);	// Base Image Number of Line (R400h)
	write_data(0x3500);		// Page 36 of SPFD5420A Datasheet
	write_command(0x0401);	// Base Image Display Control (R401h)
	write_data(0x0001);		// Page 39 of SPFD5420A Datasheet
	write_command(0x0404);	// Based Image Vertical Scroll Control (R404h)
	write_data(0x0000);		// Page 40 of SPFD5420A Datasheet

	//--------------- Normal set ---------------//
	write_command(0x0000);	// ID Read Register (R000h)
	write_data(0x0000);		// Page 13 of SPFD5420A Datasheet
	write_command(0x0001);	// Driver Output Control Register (R001h)
	write_data(0x0100);		// Page 14 of SPFD5420A Datasheet
	write_command(0x0002);	// LCD Driving Waveform Control (R002h)
	write_data(0x0100);		// Page 14 of SPFD5420A Datasheet
	write_command(0x0003);	// Entry Mode (R003h)
	write_data(0x0030);		// Page 15 of SPFD5420A Datasheet
//	write_data(0x1030);		// Page 15 of SPFD5420A Datasheet
/*	write_command(0x0006);	// Display Control 1 (R007h)
	write_data(0x0000);		// Page 16 of SPFD5420A Datasheet*/	
	write_command(0x0008);	// Display Control 2 (R008h)
	write_data(0x0808);		// Page 17 of SPFD5420A Datasheet
	write_command(0x0009);	// Display Control 3 (R009h)
	write_data(0x0001);		// Page 18 of SPFD5420A Datasheet
	write_command(0x000B);	// Low Power Control (R00Bh)
	write_data(0x0010);		// Page 19 of SPFD5420A Datasheet
	write_command(0x000C);	// External Display Interface Control 1 (R00Ch)
	write_data(0x0000);		// Page 19 of SPFD5420A Datasheet
	write_command(0x000F);	// External Display Interface Control 2 (R00Fh)
	write_data(0x0000);		// Page 20 of SPFD5420A Datasheet
	write_command(0x0007);	// Display Control 1 (R007h)
	write_data(0x0001);		// Page 16 of SPFD5420A Datasheet

	//--------------- Panel interface control 1~6 ---------------//
	write_command(0x0010);	// Panel Interface Control 1 (R010h)
	write_data(0x0012);		// Page 20 of SPFD5420A Datasheet
	write_command(0x0011);	// Panel Interface Control 2 (R011h)
	write_data(0x0202);		// Page 21 of SPFD5420A Datasheet
	write_command(0x0012);	// Panel Interface control 3 (R012h)
	write_data(0x0300);		// Page 22 of SPFD5420A Datasheet
	write_command(0x0020);	// Panel Interface control 4 (R020h)
	write_data(0x021E);		// Page 22 of SPFD5420A Datasheet
	write_command(0x0021);	// Panel Interface Control 5 (021Rh)
	write_data(0x0202);		// Page 24 of SPFD5420A Datasheet
	write_command(0x0022);	// Panel Interface Control 6 (R022h)
	write_data(0x0100);		// Page 25 of SPFD5420A Datasheet
	write_command(0x0090);	// Frame Marker Control (R090h)
	write_data(0x8000);		// Page 25 of SPFD5420A Datasheet

	//--------------- Partial display ---------------//
/*	write_command(0x0210);	// Window Horizontal RAM Address Start (R210h)
	write_data(0x0000);		// Page 35 of SPFD5420A Datasheet
	write_command(0x0211);	// Window Horziontal RAM Address End (R211h)
	write_data(0x00EF);		// Page 35 of SPFD5420A Datasheet
	write_command(0x0212);	// Window Vertical RAM Address Start (R212h)
	write_data(0x0000);		// Page 35 of SPFD5420A Datasheet
	write_command(0x0213);	// Window Vertical RAM Address End (R213h)
	write_data(0x018F);		// Page 35 of SPFD5420A Datasheet
	write_command(0x0500);	// Display Position - Partial Display 1 (R500h)
	write_data(0x0000);		// Page 40 of SPFD5420A Datasheet
	write_command(0x0501);	// RAM Address Start - Partial Display 1 (R501h)
	write_data(0x0000);		// Page 40 of SPFD5420A Datasheet
	write_command(0x0502);	// RAM Address End - Partail Display 1 (R502h)
	write_data(0x0000);		// Page 40 of SPFD5420A Datasheet
	write_command(0x0503);	// Display Position - Partial Display 2 (R503h)
	write_data(0x0000);		// Page 40 of SPFD5420A Datasheet
	write_command(0x0504);	// RAM Address Start  Partial Display 2 (R504h)
	write_data(0x0000);		// Page 41 of SPFD5420A Datasheet
	write_command(0x0505);	// RAM Address End  Partial Display 2 (R505h)
	write_data(0x0000);		// Page 41 of SPFD5420A Datasheet
	write_command(0x0606);	// Pin Control (R606h)
	write_data(0x0000);		// Page 41 of SPFD5420A Datasheet
	write_command(0x06F0);	// NVM Access Control (R6F0h)
	write_data(0x0000);		// Page 41 of SPFD5420A Datasheet*/

/*	write_command(0x0007);	// Display Control 1 (R007h)
	write_data(0x0173);		// Page 16 of SPFD5420A Datasheet
	delay(50);
	write_command(0x0007);	// Display Control 1 (R007h)
	write_data(0x0171);		// Page 16 of SPFD5420A Datasheet*/
	delay(10);
	write_command(0x0007);	// Display Control 1 (R007h)
	write_data(0x0173);		// Page 16 of SPFD5420A Datasheet

}

/*************************************************/
void SetCur(uint16 row, uint16 col)
{
	write_command(0x0201);		// GRAM Address Set (Vertical Address) (R201h)
	write_data(row);			// Page 30 of SPFD5420A Datasheet
	write_command(0x0200);		// GRAM Address Set (Horizontal Address) (R200h)
	write_data(col);			// Page 30 of SPFD5420A Datasheet

	write_command(0x0202);		// Write Data to GRAM (R202h)
}

/*************************************************/
void SetMode(uint8 arg)
{
	mode = arg;
	CTRL &= 0xF8;
	CTRL |= (mode == 16) ? MODE16 : MODE8;		// switch modes
}

/*************************************************/
void display_rgb(uint16 data)
{
	uint16 i,j;

	for(i=0;i<400;i++)
	{
		for(j=0;j<240;j++)
		{
			write_data(data);
		}
	}
}

/*************************************************/
void LCD_test()
{
	uint16 i,j;
	SetCur(0, 0);

	for(j = 0; j < 400; j++)
	{
		for(i = 0; i < 240; i++)
		{
			if (j>349) write_data(i * 4);
			else if (j>299) write_data(RED);
			else if (j>249) write_data(CYAN);
			else if (j>199) write_data(GREEN);
			else if (j>149) write_data(MAGENTA);
			else if (j>99) write_data(BLUE);
			else if (j>49) write_data(YELLOW);
			else write_data(WHITE);
		}
	}
}

/*************************************************/
void LCD_test2(void)
{
	uint16 i,j,hue;


	for(i = 210; i < 220; i++)
	{
		SetCur(i, 0);

		for(j = 0; j < 120; j++)
		{
			hue = RGB(j % 32,(j % 32) * 2,j % 32);
			write_data(hue);
			write_data(hue);
		}
	}
}

/*************************************************/
void port_test(uint8 volatile *p)
{
	char bit, rep, d;

	bit = 1;
	*p = 0;
	for (rep = 1; rep <= 8; rep++)
	{
		for (d = 0; d < rep; d++)
		{
			*p |= bit;
			DLY;
			*p &= ~bit;
		}
		bit <<= 1;
		delay(10);
	}
}

/*************************************************/
int main()
{
	uint8 a, b;
	uint16 din[16], dout[16];

	// Control
	DDR_CT = OUTS;	// all outputs
	CTRL = RD + WR + CD + CS + RES;

	// data
	DDR_HI = OUTS;		// high data port, all outputs
	DDR_LO = OUTS;		// low data port, all outputs

	// interface mode
	SetMode(16);
	
	// misc
	PORTD = 2;
	DDRD  = 2;

	delay(1000);

	// Initialize the display
	CLR_RESET;
	delay(10000);
	SET_RESET;
	delay(25000);

	init();

	SetCur(0, 0);

	//display_rgb(YELLOW);

	LCD_test();
	LCD_test2();

	write_command(0);
	LCDRead();				// dummy read
	din[0] = LCDRead();		// get chip ID


	for (a = 0; a < 16; a++)
	{
		SetCur(10 + a, 100);
		for (b = 0; b < 10; b++)
		{
			dout[a] = (1 << a);
			write_data(dout[a]);
		}
	}

		LCDRead();				// dummy read
	for (a = 0; a < 16; a++)
	{
		SetCur(10 + a, 102);
		LCDRead();				// dummy read
		din[a] = LCDRead();	// get pixel color
	}

	while (1)
	{
		PORTD ^= 2;		// blinking LED
		delay(0xFFFF);
//		port_test(&OUT_HI);
//		port_test(&OUT_LO);
//		port_test(&CTRL);
	}
	return 0;
}
